#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Diego Rubert
#
# This file is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software. If not, see <https://www.gnu.org/licenses/>


from os import path, fdopen
import sys
from argparse import ArgumentParser, FileType
import textwrap
import networkx as nx

#import pdb
#from pprint import pprint as pp

__author__ = 'diego'


def readTab(tab):
    """Read table translating gene locus to IDs."""
    return { locus: int(geneid) for geneid, locus in ( line.split() for line in tab ) }


def readABC(abc):
    """Read edges with (optional) weights in abc format."""
    edges = []
    for line in abc:
        line = line.split()
        if len(line) == 3:
            u, v, w = line
            w = float(w)
        else:
            u, v = line
            w = 1.0
        edges.append( (u, v, w) )
    return edges


def readSIF(sif):
    """Read edges with (optional) weights in sif format."""
    edges = dict()
    singletons = []
    for line in sif:
        line = line.split()
        u = line[0]
        entries = line[2:]
        if not entries:
            singletons.append(u)
        for entry in entries:
            v, *w = entry.rsplit(':', maxsplit=1)
            w = float(w[0]) if w else 1.0
            edges[frozenset((u, v))] = w
    edges = [ (*e, w) for e, w in edges.items() ]
    return edges, singletons


def familiesFromGraphByConnectedComponents(G):
    """Define families by connected components of the graph G."""
    families = [ list(c) for c in nx.connected_components(G) ]
    for f in families: # sort families
        f.sort()
    families.sort(key=len, reverse=True)
    return families


def edgeTranslateLocusTagToNumericId(edges_locus, tab):
    """Translate locus tags in edges to numeric IDs according to tab (mapping)."""
    edges = []
    absent = set()
    for u, v, w in edges_locus:
        uid = tab.get(u)
        vid = tab.get(v)
        if uid and vid:
            edges.append( (uid, vid, w) )
        else:
            if not uid:
                absent.add(u)
            if not vid:
                absent.add(v)
    return edges, absent


def vertexTranslateLocusTagToNumericId(vertices_locus, tab):
    """Translate locus tags of vertices to numeric IDs according to tab (mapping)."""
    vertices = []
    absent = set()
    for v in vertices_locus:
        vid = tab.get(v)
        if vid:
            vertices.append(vid)
        else:
            absent.add(v)
    return vertices, absent
    

def writeMCI(families, mci, comment):
    """Write families to file in mci format."""    
    maxlen = len(str(len(families))) # lenght of the highest family number
    if len(families) == 0:
        rows = 0
    else:
        rows = max(f[-1] for f in families) + 1 # works beause families are sorted
    columns = len(families)
    indent = " " * (maxlen + 3) # after first line

    mci.write("# %s\n" % comment)
    mci.write("(mclheader\nmcltype matrix\ndimensions %dx%d\n)\n(mclmatrix\nbegin\n" % (rows, columns))

    for n, family in zip(range(len(families)), families):
        text = "%-*d " % (maxlen, n)
        text += " ".join(str(geneid) for geneid in family)
        text += " $"
        text = textwrap.fill(text, width=70, subsequent_indent=indent, break_long_words=False, replace_whitespace=False, drop_whitespace=True)
        text += '\n' # must be added after fill because of drop_whitespace
        mci.write(text)
    
    mci.write(")\n")
    mci.close()

if __name__ == "__main__":
    description = 'Reads an abc or sif file with edges (optional weights) and a tab file given by mcxload, and generates an mci file with clusters given by connected components (using edges of abc or sif). DO NOT CONFUSE with the mci file with edges generated by mcxload. See https://micans.org/mcl/man/clmprotocols.html . **abc FILES DO NOT CONTAIN VERTICES WITH NO EDGES**.'
    parser = ArgumentParser(description=description)

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--abc', metavar='file.abc', type=FileType('r'), help='The abc file with edges.')
    group.add_argument('-s', '--sif', metavar='file.sif', type=FileType('r'), help='The sif file with edges.')
    
    parser.add_argument('-t', '--tab', required=True, metavar='file.tab', type=FileType('r'), help='The tab file generated by mcxload, translating gene locus tags to a numeric ID.')
    parser.add_argument('-m', '--mci', required=True, metavar='file.mci', type=FileType('w'), help='Output with clusters in mci format with genes identified by their numeric IDs.')
    parser.add_argument('-i', '--ignore-absent', action="store_true", help='Ignore genes in --abc/--sif absent in --tab (instead of generating an error).')

    
    args = parser.parse_args()

    tab = readTab(args.tab)
    if args.abc:
        edges_locus, singletons_locus = readABC(args.abc), []
    else:
        edges_locus, singletons_locus = readSIF(args.sif)

    # translate edges and vertices from locus to ID
    edges, absent_e = edgeTranslateLocusTagToNumericId(edges_locus, tab)
    singletons, absent_v = vertexTranslateLocusTagToNumericId(singletons_locus, tab)
    absent = absent_e.union(absent_v)
    
    G = nx.Graph()
    G.add_weighted_edges_from(edges)
    G.add_nodes_from(singletons)

    if len(absent) > 0 and not args.ignore_absent:
        print("ERROR: %d genes absent in %s" % (len(absent), args.tab.name), file=sys.stderr)
        exit(1)
    print("%d genes absent in %s" % (len(absent), args.tab.name))
    
    families = familiesFromGraphByConnectedComponents(G)
    writeMCI(families, args.mci, "Clusters/families given by connected components of the graph defined in edges of %s" % (args.abc or args.sif).name)
    print("Wrote %d families/clusters comprising %d genes" % (len(families), len(G)))
    print("%d edges and %d singletons (isolated vertices)" % (len(edges), len(singletons)))
    
    exit(0)

