#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Diego Rubert
#
# This file is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software. If not, see <https://www.gnu.org/licenses/>


from ffdcjid_util import readGenomesCfg, readSimilaritiesFile, FAMILY_FREE_ID, PATTERN_LOCUS_TAG, PATTERN_PROTID
from argparse import ArgumentParser, FileType, RawDescriptionHelpFormatter
from os import path
from sys import stdout, stderr

def get_locus_tag(gene):
    """Finds the locus tag of a gene by its complete ID."""
    return PATTERN_LOCUS_TAG.match(gene['id']).groups()[0]

def get_protein_id(gene):
    """Finds the protein id of a gene by its complete ID."""
    return PATTERN_PROTID.match(gene['id']).groups()[0]

def get_complete_id(gene):
    """Finds the complete ID."""
    return gene['id']#PATTERN_LOCUS_TAG.match(gene['id']).groups()[0]

def main(): # so variables are NOT global
    descr="""\
Translate numeric gene IDs to some textual identifier in a .sim file, writing to
stdout or a new file. Must also be given the genomes.cfg generated
by FFGC.
""" 
    parser = ArgumentParser(description=descr, formatter_class=RawDescriptionHelpFormatter)

    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('--use-locus-tag', dest="get_identifier", required=False, action="store_const", const=get_locus_tag,
                       help='Use locus tags to identify genes.')
    group.add_argument('--use-protein-id', dest="get_identifier", required=False, action="store_const", const=get_protein_id,
                       help='Use protein IDs as locus tags to identify genes.')
    group.add_argument('--use-complete-id', dest="get_identifier", required=False, action="store_const", const=get_complete_id,
                       help='Use complete gene IDs as locus tags to identify genes (default).')
    parser.set_defaults(get_identifier=get_complete_id)

    parser.add_argument('-C', '--cfg', required=True, type=FileType('r'), metavar='genomes.cfg',
                        help='Genomes configuration file generated by FFGC.')
    parser.add_argument('-t', '--threshold', metavar='float', type=float, default=0.0,
                        help='Threshold in interval [0,1] for edge weights (smaller are ignored, default: 0.0).')
    parser.add_argument('-i', '--sim', metavar='file.sim', required=True, type=FileType('r'),
                        help='Marker similarity files (given by FFGC), genome names are guessed from the file name ' + \
                             'in the format Gx_Gy.sim, where Gx and Gy are the genome names (as in genomes.cfg).')
    parser.add_argument('-f', '--family-filter', metavar='file.families', required=False, type=FileType('r'),
                        help='Filter out edges (pairs) connecting genes that are not in the same family, ' + \
                             'where each line of the given file contains the locus tags of a family.')
    parser.add_argument('-o', '--sim-out', metavar='file.labels.sim', required=False, type=FileType('w'), default=stdout,
                        help='Output file with labels instead of gene numeric IDs (default: stdout).')

    parser.add_argument('--omit-chromosome-names', action='store_true', required=False,
                        help='Omit columns with chromosome names (columns 1 and 3).')
    parser.add_argument('--omit-orientation', action='store_true', required=False,
                        help='Omit column with relative orientation between the gene pair (column 5).')

    args = parser.parse_args()

    print("Reading genomes configuration.", file=stderr)
    genomes = readGenomesCfg(args.cfg)

    print("Reading similarities file.", file=stderr)
    genome1, genome2 = path.splitext(path.basename(args.sim.name))[0].split('_') # genome names
    edges = readSimilaritiesFile(args.sim, FAMILY_FREE_ID, args.threshold, not args.omit_chromosome_names)
    print("%d gene similarity edges after threshold cut (threshold = %g)." % (len(edges), args.threshold), file=stderr)

    edges_locus = list()
    for e in edges:
        gene1 = genomes[genome1]['genes'][e['g1']]
        locus1 = args.get_identifier(gene1)
        gene2 = genomes[genome2]['genes'][e['g2']]
        locus2 = args.get_identifier(gene2)
        w = e['weight']

        data = (not args.omit_chromosome_names and e['chr1'],
                locus1,
                not args.omit_chromosome_names and e['chr2'],
                locus2,
                not args.omit_orientation and str(e['rel_orient']),
                str(w))
        new_e = tuple(d for d in data if d)
        edges_locus.append(new_e)
    
    if args.family_filter:
        print("Reading families and filtering.", file=stderr)
        fmap = dict()
        for line in args.family_filter:
            family = line.split()
            fmap.update( (gene, family) for gene in family )
        edges_locus = [ e for e in edges_locus if fmap.get(e[0]) == fmap.get(e[1]) != None ]
        
    print("Writing output.", file=stderr)
    for e in edges_locus:
        print('\t'.join(e), file=args.sim_out)
    
    exit(0)


if __name__ == '__main__':
    main()

